from sqlalchemy import Engine, create_engine, select
from sqlalchemy.orm import Session

import pytest

from rsqlalchemy.parse import query_to_sql
from tests.model import Base, User


@pytest.fixture(scope="module")
def get_db():
    engine = create_engine("sqlite://", echo=True)
    Base.metadata.create_all(engine)
    test_data = [
        User(name="Alice", address="123 Builder Drive"),
        User(name="Bob", address="456 Builder Avenue"),
        User(name="Charlie", address="789 Chocolate Factory Street"),
        User(name="Charlie", address="101 Flower Street"),
        User(name="Eve", address="303 River Road"),
        User(name="Frank", address="303 River Road"),
    ]
    with Session(engine) as session:
        session.add_all(test_data)
        session.commit()
    return engine


def test_eq(get_db: Engine):
    test_cases = [
        ("id==1", [1]),
        ("name==Bob", [2]),
        ("name==Charlie", [3, 4]),
        ("name==Charlie and address==*Chocolate*", [3]),
        ("name==Charlie or name==Eve or name==Frank", [3, 4, 5, 6]),
        ("name==Charlie,name==Eve,name==Frank", [3, 4, 5, 6]),
    ]
    with Session(get_db) as session:
        for query, expected in test_cases:
            result = session.scalars(
                select(User.id).where(query_to_sql(User, query))
            ).all()
            assert result == expected

def test_in(get_db: Engine):
    test_cases = [
        ("id=in=(1,2,3)", [1,2,3]),
        ("id=in=(1)", [1]),
        ("id=in=()", []),
        ("id=in=(1,2,3) or name=in=(Charlie,Eve,Frank)", [1,2,3,4,5,6]),
        ("id=in=(1,2,3) and name=in=(Charlie,Eve,Frank)", [3]),
    ]
    with Session(get_db) as session:
        for query, expected in test_cases:
            result = session.scalars(
                select(User.id).where(query_to_sql(User, query))
            ).all()
            assert result == expected



from sqlalchemy import ColumnElement, and_, or_
from sqlalchemy.orm import DeclarativeBase

from rsqlalchemy.utils import split_query, is_group


def query_to_sql(orm_model: DeclarativeBase, s: str) -> ColumnElement:
    return expression_or(orm_model, s)


def expression_or(orm_model: DeclarativeBase, s: str) -> ColumnElement:
    children = split_query(s, [" or ", ","])
    parsed_children = []
    if len(children) <= 1:
        return expression_and(orm_model, s)
    for child in children:
        if is_group(child):
            parsed_children.append(expression_or(orm_model, child[1:-1]))
        else:
            parsed_children.append(expression_and(orm_model, child))
    return or_(*parsed_children)


def expression_and(orm_model: DeclarativeBase, s: str) -> ColumnElement:
    children = split_query(s, [" and ", ";"])
    parsed_children = []
    if len(children) <= 1:
        return comparison(orm_model, s)
    for child in children:
        if is_group(child):
            parsed_children.append(expression_or(orm_model, child[1:-1]))
        else:
            parsed_children.append(comparison(orm_model, child))
    return and_(*parsed_children)


def comparison(orm_model: DeclarativeBase, s: str) -> ColumnElement:
    for comparison_op in ["==", "=in=", "=out=", "<=", ">="]:
        if comparison_op in s:
            selector, arguments = s.split(comparison_op)
            break
    else:
        raise ValueError(f"{s} is not a supported comparison")

    column: ColumnElement = getattr(orm_model, selector)

    if is_group(arguments):
        args = arguments.strip("()").split(",")
        args = [None if x == "null" else x for x in args]
        if comparison_op == "=in=":
            return column.in_(args)
        elif comparison_op == "=out=":
            return column.notin_(args)
        else:
            raise ValueError(f"{comparison_op} cannot be used with group arguments")

    if arguments == "null":
        assert comparison_op == "==", "null cannot be used with other operators"
        arguments = None

    if comparison_op == "==":
        if arguments and "*" in arguments:
            return column.like(arguments.replace("*", "%"))
        else:
            return column == arguments
    elif comparison_op == "<=":
        return column <= arguments
    elif comparison_op == ">=":
        return column >= arguments
    else:
        raise ValueError(f"{comparison_op} cannot be used with group arguments")



from rsqlalchemy import utils


def test_split_ignore_parenth():
    test_cases = [
        ["a and (b and c or d) and e", [" and "], ["a", "(b and c or d)", "e"]],
        ["(a or b or c)", [" or "], ["(a or b or c)"]],
        ["a ( ) b", [" "], ["a", "( )", "b"]],
        ["split", ["split"], []],
        ["a and (b;(c));d", [" and ", ";"], ["a", "(b;(c))", "d"]],
        [
            "a,(b),c,(d,e),f,(g,(h,i))",
            [","],
            ["a", "(b)", "c", "(d,e)", "f", "(g,(h,i))"],
        ],
    ]
    for target, delimiter, expected in test_cases:
        actual = utils.split_query(target, delimiter)
        assert (
            expected == actual
        ), f"\nTarget: {target}\nExpected: {expected}\nActual: {actual}"


def test_split_ignore_parenth_error():
    try:
        utils.split_query("a or b or c)))))", [" or "])
    except ValueError as e:
        assert e.args == ("Invalid expression: too many right parentheses",)
    else:
        assert False, "Test should raise ValueError but didn't"
    try:
        utils.split_query("(((a or b or c", [" or "])
    except ValueError as e:
        assert e.args == ("Invalid expression: too many left parentheses",)
    else:
        assert False, "Test should raise ValueError but didn't"



from typing import Optional
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(DeclarativeBase):
    pass


class User(Base):
    __tablename__ = "user"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(64))
    address: Mapped[Optional[str]]



from sqlalchemy import ColumnElement, and_, or_

from rsqlalchemy.parse import comparison, expression_and, expression_or
from tests.model import User


def _compile(expression: ColumnElement) -> str:
    return str(expression.compile(compile_kwargs={"literal_binds": True}))


def test_comparison():
    test_cases = [
        ("id==1", User.id == "1"),
        ("id>=1", User.id >= "1"),
        ("id<=10", User.id <= "10"),
        ("name=in=(a,b,c)", User.name.in_(["a", "b", "c"])),
        ("address==null", User.address.is_(None)),
        ("name==*abc", User.name.like("%abc")),
        ("name==*abc*", User.name.like("%abc%")),
        ("name==ab*c", User.name.like("ab%c")),
        (
            "address=out=(a,b,c)",
            User.address.notin_(["a", "b", "c"]),
        ),
        (
            "address=in=(null,a,b)",
            User.address.in_([None, "a", "b"]),
        ),
    ]
    for query, expected in test_cases:
        assert _compile(comparison(User, query)) == _compile(expected)


def test_expression_and():
    test_cases = [
        ("id==1;id==2", and_(User.id == "1", User.id == "2")),
        ("id==1 and id==2", and_(User.id == "1", User.id == "2")),
        ("id==1 and id==2;id==3", and_(User.id == "1", User.id == "2", User.id == "3")),
        (
            "id==1 and (id==2;id==3)",
            and_(User.id == "1", and_(User.id == "2", User.id == "3")),
        ),
        (
            "(id>=1 and name=in=(a,b,c)) and (id==2;id==3)",
            and_(
                and_(User.id >= "1", User.name.in_(["a", "b", "c"])),
                and_(User.id == "2", User.id == "3"),
            ),
        ),
    ]
    for query, expected in test_cases:
        assert _compile(expression_and(User, query)) == _compile(expected)


def test_expression_or():
    test_cases = [
        ("id==1,id==2", or_(User.id == "1", User.id == "2")),
        ("id==1 or id==2", or_(User.id == "1", User.id == "2")),
        ("id==1 or id==2,id==3", or_(User.id == "1", User.id == "2", User.id == "3")),
        (
            "id==1 or id==2 and id==3",
            or_(
                User.id == "1",
                and_(User.id == "2", User.id == "3"),
            ),
        ),
        (
            "(id==1 or id==2) and id==3",
            and_(
                or_(User.id == "1", User.id == "2"),
                User.id == "3",
            ),
        ),
        (
            "(id==1 or id==2) and id==3",
            and_(
                or_(User.id == "1", User.id == "2"),
                User.id == "3",
            ),
        ),
        (
            "(id==1 or id==2 and id==3) and id==4",
            and_(
                or_(User.id == "1", and_(User.id == "2", User.id == "3")),
                User.id == "4",
            ),
        ),
        (
            "((id==1 or id==2) and id==3) and id==4",
            and_(
                and_(or_(User.id == "1", User.id == "2"), User.id == "3"),
                User.id == "4",
            ),
        ),
    ]
    for query, expected in test_cases:
        assert _compile(expression_or(User, query)) == _compile(expected)



import re


def is_group(s: str) -> bool:
    return s.startswith("(") and s.endswith(")")


def split_query(target: str, delimiters: list[str]) -> list[str]:
    level = 0
    preserved: list[str] = []
    left = 0
    marker = "\uffff"
    marked = ""

    # find all substring enclosed in parenthesis
    for right in range(len(target)):
        if target[right] == "(":
            if level == 0:
                left = right
                marked += marker
            level += 1
        elif target[right] == ")":
            level -= 1
            if level == 0:
                preserved.append(target[left : right + 1])
        if level < 0:
            raise ValueError("Invalid expression: too many right parentheses")
        elif level == 0 and target[right] != ")":
            marked += target[right]
    if level > 0:
        raise ValueError("Invalid expression: too many left parentheses")

    # split and put content inside parentheses back in place
    result = []
    regex_pattern = "|".join(map(re.escape, delimiters))
    for word in re.split(regex_pattern, marked):
        if word == "":
            continue
        elif marker in word:
            result.append(word.replace(marker, preserved.pop(0)))
        else:
            result.append(word)

    return result
